from __future__ import annotations

import asyncio
import logging

import aiohttp
from aiohttp import client_exceptions

from .clip import Clip
from .config import Config
from .errors import raise_error
from .groups import Groups
from .lights import Lights
from .scenes import Scenes
from .sensors import Sensors

_DEFAULT = object()


class Bridge:
    """Control a Hue bridge."""

    def __init__(
        self,
        host: str,
        websession: aiohttp.ClientSession,
        *,
        username: str | None = None,
        bridge_id: str | None = None,
    ):
        self.host = host
        self.username = username
        self.websession = websession
        self._bridge_id = bridge_id

        self.proto = None
        self.config = None
        self.groups = None
        self.lights = None
        self.scenes = None
        self.sensors = None
        self.clip = Clip(self.request_2)

        self.logger = logging.getLogger(f"{__name__}.{host}")

        # self.capabilities = None
        # self.rules = None
        # self.schedules = None

    @property
    def id(self):
        """Return the ID of the bridge."""
        if self.config is not None:
            return self.config.bridgeid

        return self._bridge_id

    async def create_user(self, device_type):
        """Create a user.

        https://developers.meethue.com/documentation/configuration-api#71_create_user
        """
        result = await self.request("post", "", {"devicetype": device_type}, auth=False)
        self.username = result[0]["success"]["username"]
        return self.username

    async def initialize(self):
        result = await self.request("get", "")

        self.config = Config(result.pop("config"), self.request)
        self.groups = Groups(self.logger, result.pop("groups"), self.request)
        self.lights = Lights(self.logger, result.pop("lights"), self.request)
        if "scenes" in result:
            self.scenes = Scenes(self.logger, result.pop("scenes"), self.request)
        if "sensors" in result:
            self.sensors = Sensors(self.logger, result.pop("sensors"), self.request)

        self.logger.debug("Unused result: %s", result)

    async def request(self, method, path, json=None, auth=True):
        """Make a request to the API."""
        # By default we assume we need to connect over `https`
        # Old bridges and incompatible emulates still use `http` so we force a fallback
        # We will store protocol in `self.proto` if request succesful.
        if self.proto is None:
            proto = "https"
        else:
            proto = self.proto

        url = "{}://{}/api/".format(proto, self.host)
        if auth:
            url += "{}/".format(self.username)
        url += path

        try:
            async with self.websession.request(
                method, url, json=json, ssl=False
            ) as res:
                res.raise_for_status()

                # Store the protocol that worked
                if self.proto is None:
                    self.proto = proto

                data = await res.json()
                _raise_on_error(data)
                return data

        except client_exceptions.ClientConnectionError:
            if self.proto is not None:
                raise

            self.proto = "http"
            return await self.request(method, path, json, auth)

    async def request_2(self, method, path, timeout=_DEFAULT):
        """Make a request to any path with Hue's new request method.

        This method has the auth in a header.
        """
        url = f"{self.proto or 'https'}://{self.host}/{path}"

        kwargs = {
            "ssl": False,
            "headers": {"hue-application-key": self.username},
        }

        if timeout is not _DEFAULT:
            kwargs["timeout"] = timeout

        async with self.websession.request(method, url, **kwargs) as res:
            res.raise_for_status()
            return await res.json()

    async def listen_events(self):
        """Listen to events and apply changes to objects."""
        pending_events = asyncio.Queue()

        async def receive_events():
            while True:
                self.logger.debug("Subscribing to events")
                try:
                    for event in await self.clip.next_events():
                        self.logger.debug("Received event: %s", event)
                        pending_events.put_nowait(event)
                except client_exceptions.ServerDisconnectedError:
                    self.logger.debug("Event endpoint disconnected")
                except client_exceptions.ClientError as err:
                    if isinstance(err, client_exceptions.ClientResponseError):
                        # We get 503 when it's too busy, but any other error
                        # is probably also because too busy.
                        self.logger.debug(
                            "Got status %s from endpoint. Sleeping while waiting to resolve",
                            err.status,
                        )
                    else:
                        self.logger.debug("Unable to reach event endpoint: %s", err)

                    await asyncio.sleep(5)
                except asyncio.TimeoutError:
                    pass
                except Exception:
                    self.logger.exception("Unexpected error")
                    pending_events.put(None)
                    break

        event_task = asyncio.create_task(receive_events())

        while True:
            try:
                event = await pending_events.get()
            except asyncio.CancelledError:
                event_task.cancel()
                await event_task
                raise

            # If unexpected error occurred
            if event is None:
                return

            if event["type"] not in ("update", "motion"):
                self.logger.debug("Unknown event type: %s", event)
                continue

            for event_data in event["data"]:
                # We don't track object that groups all items (bridge_home)
                if event_data["id_v1"] == "/groups/0":
                    continue

                item_type = event_data["id_v1"].split("/", 2)[1]

                if item_type not in (
                    # These all inherit from APIItems and so can handle events
                    "lights",
                    "sensors",
                    "scenes",
                    "groups",
                ):
                    self.logger.debug(
                        "Received %s event for unknown item type %s: %s",
                        event["type"],
                        item_type,
                        event_data,
                    )
                    continue

                obj = getattr(self, item_type).process_event(event["type"], event_data)
                # if obj is None, we didn't know the object
                # We could consider triggering a full refresh
                if obj is not None:
                    yield obj


def _raise_on_error(data):
    """Check response for error message."""
    if isinstance(data, list):
        data = data[0]

    if isinstance(data, dict) and "error" in data:
        raise_error(data["error"])



class APIItems:
    """Base class for a map of API Items."""

    def __init__(self, logger, raw, request, path, item_cls):
        self._logger = logger
        self._request = request
        self._path = path
        self._item_cls = item_cls
        self._items = {}
        self._process_raw(raw)

    async def update(self):
        raw = await self._request("get", self._path)
        self._process_raw(raw)

    def _process_raw(self, raw):
        for id, raw_item in raw.items():
            obj = self._items.get(id)

            if obj is not None:
                obj.raw = raw_item
            else:
                self._items[id] = self._item_cls(id, raw_item, self._request)

        removed_items = []

        for id in self._items:
            if id not in raw:
                removed_items.append(id)

        for id in removed_items:
            del self._items[id]

    def values(self):
        return self._items.values()

    def process_event(self, event_type: str, data: dict):
        id_v1 = data["id_v1"]
        obj_id = id_v1.rsplit("/", 1)[1]
        obj = self._items.get(obj_id)

        if obj is None:
            self._logger.debug(
                "Received %s event for unknown item %s: %s", event_type, id_v1, data
            )
            return None

        meth = getattr(obj, f"process_{event_type}_event", None)

        if meth is None:
            self._logger.debug(
                "Don't know how to handle %s event for %s (%s): %s",
                event_type,
                id_v1,
                type(obj).__name__,
                data,
            )
            return None

        meth(data)
        return obj

    def __getitem__(self, obj_id):
        return self._items[obj_id]

    def __iter__(self):
        return iter(self._items)



"""Aiohue errors.

https://developers.meethue.com/documentation/error-messages
"""


class AiohueException(Exception):
    """Base error for aiohue."""


class Unauthorized(AiohueException):
    """Username is not authorized."""


class LinkButtonNotPressed(AiohueException):
    """Raised when trying to create a user but link button not pressed."""


ERRORS = {1: Unauthorized, 101: LinkButtonNotPressed}


def raise_error(error):
    type = error["type"]
    cls = ERRORS.get(type, AiohueException)
    raise cls("{}: {}".format(type, error["description"]))



from .bridge import Bridge
from .util import normalize_bridge_id

URL_NUPNP = "https://discovery.meethue.com/"


async def discover_nupnp(websession):
    """Discover bridges via NUPNP."""
    async with websession.get(URL_NUPNP) as res:
        return [
            Bridge(
                item["internalipaddress"],
                bridge_id=normalize_bridge_id(item["id"]),
                websession=websession,
            )
            for item in (await res.json())
        ]



"""Utils for aiohue."""
import logging


def normalize_bridge_id(bridge_id: str):
    """Normalize a bridge identifier."""
    bridge_id = bridge_id.lower()

    # zeroconf: properties['id'], field contains semicolons after each 2 char
    if len(bridge_id) == 17 and sum(True for c in "aa:bb:cc:dd:ee:ff" if c == ":"):
        return bridge_id.replace(":", "")

    # nupnp: contains 4 extra characters in the middle: "fffe"
    if len(bridge_id) == 16 and bridge_id[6:10] == "fffe":
        return bridge_id[0:6] + bridge_id[-6:]

    # SSDP/UPNP and Hue Bridge API contains right ID.
    if len(bridge_id) == 12:
        return bridge_id

    logging.getLogger(__name__).warn("Received unexpected bridge id: %s", bridge_id)

    return bridge_id



class Clip:
    """Represent Hue clip."""

    def __init__(self, request_2):
        self._request_2 = request_2

    async def next_events(self):
        """Note, this method will be pending until next event."""
        return await self._request_2("get", "eventstream/clip/v2", timeout=360)

    async def resources(self):
        """Fetch resources from Hue.

        Available types:

        homekit
        device
        bridge
        zigbee_connectivity
        entertainment
        light
        bridge_home
        grouped_light
        room
        scene
        """
        return await self._request_2("get", "clip/v2/resource")



class Config:
    """Represent Hue config.

    https://developers.meethue.com/documentation/configuration-api#72_get_configuration
    """

    def __init__(self, raw, request):
        self.raw = raw
        self._request = request

    @property
    def name(self):
        """Name of the bridge."""
        return self.raw["name"]

    @property
    def swversion(self):
        """Software version of the bridge."""
        return self.raw["swversion"]

    @property
    def swupdate2_bridge_state(self):
        """Software update state of the bridge."""
        return self.raw.get("swupdate2", {}).get("bridge", {}).get("state")

    @property
    def modelid(self):
        """Model ID of the bridge."""
        return self.raw["modelid"]

    @property
    def bridgeid(self):
        """ID of the bridge."""
        return self.raw["bridgeid"]

    @property
    def apiversion(self):
        """Supported API version of the bridge."""
        return self.raw["apiversion"]

    @property
    def mac(self):
        """Mac address of the bridge."""
        return self.raw["mac"]

    async def update(self):
        self.raw = await self._request("get", "config")
